{% extends "base.html" %}

{% block extra_resources %}
<link rel="stylesheet" href="statics/css/status.css">
{% endblock %}

{% block content %}
{% from macros import embed %}
<div id="status-fluid">

  <div id="status-nbytes-cluster">
    {{ embed(roots.nbytes_cluster) }}
  </div>

  <div id="status-nbytes-workers">
    {{ embed(roots.nbytes_workers) }}
  </div>

  <div id="status-processing">
    {{ embed(roots.processing) }}
  </div>

  <div id="status-tasks">
    {{ embed(roots.task_stream) }}
  </div>

  <div id="status-progress">
    {{ embed(roots.task_progress) }}
  </div>

</div>
{{ plot_script }}

{% endblock %}
